#include "cpgl.h"
#include "obj.h"
#include "model.h"
#include "camera.h"
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <vector>

//---------------------------------------------
typedef struct WaveSrc_s {
    float x;          // source x
    float z;          // source y
    float amp_0;      // initial amplitude  [0,.2]
    float lambda_0;   // initial wavelength [5,10]
    float start_time; // initial start time
    float speed;     
} WaveSrc;

#define AMP_RANGE    .3f
#define AMP_MIN      .1f
#define LAMBDA_RANGE 1.0f
#define LAMBDA_MIN   0.75f
#define SPEED_MIN    5.0f
#define SPEED_RANGE  1.0f

std::vector< WaveSrc > wave_vec;
//--------------------------------------------

#ifndef PI
#define PI 3.14159
#endif

#define DEG2RAD( x ) ( (x) / 180.0f * PI )
#define RAD2DEG( x ) ( (x) / PI * 180.0f )

//-----------------------------------
SDL_Surface* gDrawSurface = NULL;
int  width     = 800;
int  height    = 600;
bool BreakLoop = false;
bool Capture   = false;

//-----------------------------------
// Update Function Pointer and Variables Updated Each Call
void (*updatefn)(double) = NULL;
float   TotalTime   = 0;
GLfloat LightPos[4] = { 0, 15, 5, 0};
GLfloat LightVel[3] = { 0, 0, 0 };
GLfloat CamVel[3]   = { 0, 0, 0 };
GLfloat CamPitch    = 0;
GLfloat CamYaw      = 0;
float   NMapScale   = .04;
float   NMapBias    = .03;


//----------------------------------
#define REFLECT 0  // Reflection buffer
#define REFRACT 1  // Refraction buffer
#define NRMLMAP 2  // Pool Normal Map with hight in alpha channel
#define DUDVMAP 3  // Pool DU/DV Map

GLuint  PoolFBO[4] = { 0, 0 };        // Pool FBOs
GLuint  PoolTex[4] = { 0, 0 };        // Pool FBO Texture Color Attachments
GLuint  PoolDB[2]  = { 0, 0 };        // Pool FBO Renderbuffer Depth Buffers

GLuint  FBOFrontBuffer = 0;           // Pseudo FrontBuffer for an FBO
                                      // So that way we can refer to it
                                      // if we want to do multiple passes
                                      // with the current FBO and refer
                                      // to the previous pass (GPGPU stuff)
                                      // ASSUMED TO BE RGBA!!

GLfloat UVOffset[2]    = { 0.0, 0.0 }; // Offset of UV coordinates
#define POOL_SIZE      11.0f
#define POOLBUF_SIZE   256

GLhandleARB PoolPrg;
GLhandleARB PoolVs;
GLhandleARB PoolFs;

GLhandleARB WaveHeightPrg;
GLhandleARB WaveHeightVs;
GLhandleARB WaveHeightFs;

GLhandleARB WaveNormalsPrg;
GLhandleARB WaveNormalsVs;
GLhandleARB WaveNormalsFs;

// Modes of Wave shader
#define CALC_HIEGHTS   0
#define CALC_NORMALS   1

GLhandleARB DuDvPrg;
GLhandleARB DuDvVs;
GLhandleARB DuDvFs;

GLfloat SpeedFac = .1f;
int NumWaves = 0;

//----------------------------------
// Model objects
int poolRoomObj;
int skySphereObj;
int poolBoxObj;
int poolSurfObj;
int spikedTeapotObj;

bool underwater = false;

//---------------------------------
// App specific functions
void CreatePoolFBS();
void DrawToReflectionFB();
void DrawToRefractionFB();
void CalcNormalMap();
void CalcDuDvMap();
void DrawToDefaultFB();
void PushBackWave();
void PopBackWave();
void SwapFBOBuffer(GLuint*); // swaps with FBOFrontBuffer giving you the
                             // Texture Id to the new back buffer

#if !defined(__APPLE__) && !defined(_WIN32)

// SHADER FUNCTIONS
PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectARB = NULL;
PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectARB = NULL;
PFNGLSHADERSOURCEARBPROC            glShaderSourceARB = NULL;
PFNGLCOMPILESHADERARBPROC           glCompileShaderARB = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivARB = NULL;
PFNGLATTACHOBJECTARBPROC            glAttachObjectARB = NULL;
PFNGLGETINFOLOGARBPROC              glGetInfoLogARB = NULL;
PFNGLLINKPROGRAMARBPROC             glLinkProgramARB = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectARB = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationARB = NULL;
PFNGLUNIFORM1FARBPROC               glUniform1f = NULL;
PFNGLUNIFORM2FARBPROC               glUniform2f = NULL;
PFNGLUNIFORM3FARBPROC               glUniform3f = NULL;
PFNGLUNIFORM1FVARBPROC              glUniform1fv = NULL;
PFNGLUNIFORM2FVARBPROC              glUniform2fv = NULL;
PFNGLUNIFORM3FVARBPROC              glUniform3fv = NULL;
PFNGLUNIFORM1IARBPROC               glUniform1i = NULL;

PFNGLDELETEOBJECTARBPROC            glDeleteObjectARB= NULL;
PFNGLBINDATTRIBLOCATIONARBPROC      glBindAttribLocationARB= NULL;
PFNGLGETATTRIBLOCATIONARBPROC       glGetAttribLocationARB= NULL;
PFNGLVERTEXATTRIB3FARBPROC          glVertexAttrib3fARB= NULL;

// FBO functions
PFNGLGENFRAMEBUFFERSEXTPROC         glGenFramebuffersEXT= NULL;
PFNGLBINDFRAMEBUFFEREXTPROC         glBindFramebufferEXT= NULL;
PFNGLDELETEFRAMEBUFFERSEXTPROC      glDeleteFramebuffersEXT= NULL;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC    glFramebufferTexture2DEXT= NULL;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC  glCheckFramebufferStatusEXT= NULL;

// MULTITEXTURE FUNCTIONS
PFNGLACTIVETEXTUREPROC              glActiveTexture = NULL;

#endif


//----------------------------------------------------------------
//
//----------------------------------------------------------------
int main(int argc, char** argv)
{

    printf("loading models!\n");

    SetupSDL();
    SetupGLProcs();
    InitLight();
    CreatePoolFBS();

    poolRoomObj  = load_model("models/poolroom_v2.obj",
                             "shaders/normalmap.vert",
                             "shaders/normalmap.frag");
    skySphereObj = load_model("models/sky.obj", NULL, NULL);
    spikedTeapotObj = load_model("models/spiked_teapot.obj", poolRoomObj);
    position_model(spikedTeapotObj, 3.0, 3.0, 3.0 );

    orient_model(skySphereObj, DEG2RAD(90.0), 0.0, 0.0);

    poolBoxObj   = load_model("models/poolbox.obj", poolRoomObj);

    glUniform1i( get_model_uniform_loc(poolRoomObj,"nmap"), 1);
    glUniform1i( get_model_uniform_loc(poolRoomObj,"cmap"), 0);
    glUniform1f( get_model_uniform_loc(poolRoomObj,"scale"), NMapScale);
    glUniform1f( get_model_uniform_loc(poolRoomObj,"bias"), NMapBias);
    glUniform1i( get_model_uniform_loc(poolRoomObj,"DoLighting"), 1);
    glUniform1i( get_model_uniform_loc(poolRoomObj,"apply_parallax"), 1);

    position_camera(3, 3 ,0);
    orient_camera(0,DEG2RAD(45),0);

    // shader to calculate wave height
    SetupShaders("shaders/waves_height.vert", "shaders/waves_height.frag", 
                 WaveHeightVs, WaveHeightFs, WaveHeightPrg);
    glUniform1i( glGetUniformLocationARB(WaveHeightPrg,"PrevItr"), 0);

    // shader to calculate tangent and binormal from height map
    SetupShaders("shaders/dudv.vert", "shaders/dudv.frag", DuDvVs, DuDvFs,
                 DuDvPrg);
    glUniform1f( glGetUniformLocationARB(DuDvPrg,"StepSize"), 
                 1.0f / float(POOLBUF_SIZE));
    glUniform1i( glGetUniformLocationARB(DuDvPrg,"NormalMapTex"), 0);

    // shader to calculate surface normal from dudv map
    SetupShaders("shaders/waves_normals.vert", "shaders/waves_normals.frag", 
                 WaveNormalsVs, WaveNormalsFs, WaveNormalsPrg);
    glUniform1f( glGetUniformLocationARB(WaveNormalsPrg,"StepSize"), 
                 1.0f / float(POOLBUF_SIZE));
    glUniform1i( glGetUniformLocationARB(WaveNormalsPrg,"PrevItr"), 0);

    // shader to take normal/height, dudv and apply reflection & refraction
    SetupShaders("shaders/pool.vert", "shaders/pool.frag", PoolVs, PoolFs,
                 PoolPrg);
    poolSurfObj = -1;
    glUniform1i( glGetUniformLocationARB(PoolPrg,"ReflectTex"), 0);
    glUniform1i( glGetUniformLocationARB(PoolPrg,"RefractTex"), 1);
    glUniform1i( glGetUniformLocationARB(PoolPrg,"NormalMapTex"), 2);
    glUniform1i( glGetUniformLocationARB(PoolPrg,"DuDvTex"), 3);
    glUniform1i( glGetUniformLocationARB(PoolPrg,"DepthTex"), 4);

    PushBackWave();

    updatefn = Update;
    InfLoop();
    
    // Delete Shaders
    UnloadShaders(PoolPrg,PoolVs,PoolFs);
    UnloadShaders(WaveHeightPrg,WaveHeightVs,WaveHeightFs);
    UnloadShaders(WaveNormalsPrg,WaveNormalsVs,WaveNormalsFs);
    UnloadShaders(DuDvPrg,DuDvVs,DuDvFs);

    free_all_models();

    return 0;
}

//----------------------------------------------------------------
// Load Shader Source From File
//----------------------------------------------------------------
char* GetShaderSource(const char* filename)
{
    char* source;
    FILE* fptr = fopen(filename,"rb");
    if( !fptr )
        return NULL;
    fseek(fptr,0,SEEK_END);
    long length = ftell(fptr);
    source = new char[length + 1];
    memset(source,0,length + 1);
    fseek(fptr,0,SEEK_SET);
    fread(source,sizeof(char),length,fptr);
    fclose(fptr);

    return source;
}

//----------------------------------------------------------------
//
//----------------------------------------------------------------
void UnloadShaders(GLhandleARB& vs, GLhandleARB& fs, GLhandleARB& prg)
{
    glDeleteObjectARB(prg);
}

//----------------------------------------------------------------
// Setup shaders
//----------------------------------------------------------------
void SetupShaders(const char* vert, const char* frag, GLhandleARB& vs, 
                  GLhandleARB& fs, GLhandleARB& prg)
{
    printf("======================\n");
    printf("Setting up Shaders:\nVertex Shader %s\nFragment Shader %s\n",
            vert, frag);
    char infolog[1024];
    GLint length;
    // get shader source code
    char* vertsource = GetShaderSource(vert);
    char* fragsource = GetShaderSource(frag);
    if(!vertsource && !fragsource )
    {
        printf("No shader source code!\n");
        return;
    }

    prg = glCreateProgramObjectARB();
    
    // process the vertex shader source
    if( vertsource )
    {    
        vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
        glShaderSourceARB(vs,1,(const GLcharARB**)&vertsource,NULL);
        glCompileShaderARB(vs);
       
        // see if it compiled
        GLint compiled;
        glGetObjectParameterivARB(vs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,vs);
        else
        {
            printf("VERTEX SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(vs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }
   
    // process the fragment shader source code
    if( fragsource )
    {
        fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER);
        glShaderSourceARB(fs,1,(const GLcharARB**)&fragsource,NULL);
        glCompileShaderARB(fs);
       
        // see if the fragment shader compiled
        GLint compiled;
        glGetObjectParameterivARB(fs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,fs);
        else
        {
            printf("FRAG SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(fs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }

    // bind 6th vertex attribute to attribute variable name "Tangent"
    glBindAttribLocationARB(prg, 6, "Tangent");
    
    glLinkProgramARB(prg);
    
    glGetInfoLogARB(prg,1024,&length,infolog);
    if( length )
    {
        printf("GLSL PROGRAM LOG: %s\n", infolog );
    }
    
    glUseProgramObjectARB(prg);

    printf("----- DONE ------\n");
    if( fragsource ) delete [] fragsource;
    if( vertsource ) delete [] vertsource;

}

//----------------------------------------------------------------
// Draw
//----------------------------------------------------------------
void DrawScene()
{
    glActiveTexture(GL_TEXTURE4_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE3_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE2_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    CalcNormalMap();
    CalcDuDvMap();
    DrawToReflectionFB();
    DrawToRefractionFB();
    DrawToDefaultFB();
}

//----------------------------------------------------------------
// Calculating Alpha of Normal Map for Height of Water
//----------------------------------------------------------------
void CalcNormalMap()
{    
    // ===================NORMAL MAP BUFFER : Calc Heights Mode =============//
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, PoolFBO[NRMLMAP]);
    glViewport(0,0,POOLBUF_SIZE,POOLBUF_SIZE);
    glClearColor(0.5f, 0.5f, 1.0f, 0.5f);
    glClear(GL_COLOR_BUFFER_BIT);

    glUseProgramObjectARB(0);

    SwapFBOBuffer( &(PoolTex[NRMLMAP]) );
    glClearColor(0.5f, 0.5f, 1.0f, 0.5f);

    glUseProgramObjectARB( WaveHeightPrg );
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-POOL_SIZE, POOL_SIZE, POOL_SIZE, -POOL_SIZE, 0, 1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    gluLookAt(0,0,0, 0, 0, -1, 0, 1, 0);
    
    // update some shader parameters

    // determine heights from aggregating wave heights
    int i;
    for( i = 0; i < wave_vec.size(); i++)
    {
        SwapFBOBuffer( &(PoolTex[NRMLMAP]) );
        glActiveTexture(GL_TEXTURE0_ARB);
        glBindTexture(GL_TEXTURE_2D, FBOFrontBuffer);

        glUniform2f( glGetUniformLocationARB( WaveHeightPrg, "Src"),
                     wave_vec[i].x, wave_vec[i].z);
        glUniform1f( glGetUniformLocationARB( WaveHeightPrg, "Amp"),
                     wave_vec[i].amp_0);
        glUniform1f( glGetUniformLocationARB( WaveHeightPrg, "StartTime"),
                     wave_vec[i].start_time);
        glUniform1f( glGetUniformLocationARB( WaveHeightPrg, "Frequency"),
                     2.0 * PI / wave_vec[i].lambda_0);
        glUniform1f( glGetUniformLocationARB( WaveHeightPrg, "Speed"),
                     wave_vec[i].speed);
        glUniform1f( glGetUniformLocationARB( WaveHeightPrg, "Time"), 
            TotalTime);

        glBegin(GL_QUADS);
        glTexCoord2f( 0,            1          );
        glVertex2f  ( -POOL_SIZE,   -POOL_SIZE );

        glTexCoord2f( 0,            0          );
        glVertex2f  ( -POOL_SIZE,   POOL_SIZE  );

        glTexCoord2f( 1,            0          );
        glVertex2f  ( POOL_SIZE,    POOL_SIZE  );

        glTexCoord2f( 1,            1          );
        glVertex2f  ( POOL_SIZE,    -POOL_SIZE );
        glEnd();
       
        if(Capture)
        {
            char filename[128];
            sprintf(filename,"wave_%d_fbo.png", i);
            glBindTexture(GL_TEXTURE_2D, PoolTex[NRMLMAP]);
            snap(filename,POOLBUF_SIZE, POOLBUF_SIZE,TEXBUF_SRC);
        }

    }
    if( i != NumWaves ) 
        printf("Weird things!!! Didn't draw calc all waves\n");
/*
   if(Capture)
   {
       glBindTexture(GL_TEXTURE_2D, PoolTex[NRMLMAP]);
       snap("waves_fbo.png",POOLBUF_SIZE, POOLBUF_SIZE,TEXBUF_SRC);
   }
*/
    if( wave_vec.size() % 2 )
        SwapFBOBuffer( &(PoolTex[NRMLMAP]) );

    // =============== NORMAL MAP BUFFER : Calc Normals Mode ================//
    if( wave_vec.size() > 0 )
    {
        // run one more time to calculate the normals 
        // based off the final heights
        glBindTexture(GL_TEXTURE_2D, FBOFrontBuffer); 
        glUseProgramObjectARB(WaveNormalsPrg);

        glBegin(GL_QUADS);
        glTexCoord2f( 0,            1          );
        glVertex2f  ( -POOL_SIZE,   -POOL_SIZE );
    
        glTexCoord2f( 0,            0          );
        glVertex2f  ( -POOL_SIZE,   POOL_SIZE  );
    
        glTexCoord2f( 1,            0          );
        glVertex2f  ( POOL_SIZE,    POOL_SIZE  );
    
        glTexCoord2f( 1,            1          );
        glVertex2f  ( POOL_SIZE,    -POOL_SIZE );
        glEnd();
    }

    if( Capture )
    {
        glBindTexture(GL_TEXTURE_2D, PoolTex[NRMLMAP]);
        snap("nmap_fbo.png",POOLBUF_SIZE, POOLBUF_SIZE, TEXBUF_SRC);
    }

    glPopMatrix();
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void CalcDuDvMap()
{
    // ========================== DUDV BUFFER ===============================//
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, PoolFBO[DUDVMAP]);
    glViewport(0,0,POOLBUF_SIZE,POOLBUF_SIZE);
    glClearColor(.5f, 0.5f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    
    glUseProgramObjectARB( DuDvPrg );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-POOL_SIZE, POOL_SIZE, POOL_SIZE, -POOL_SIZE, 0, 1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    gluLookAt(0,0,0, 0, 0, -1, 0, 1, 0);

    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolTex[NRMLMAP]);

    // determine change in heights
    glBegin(GL_QUADS);
    glTexCoord2f( 0,            1          );
    glVertex2f  ( -POOL_SIZE,   -POOL_SIZE );

    glTexCoord2f( 0,            0          );
    glVertex2f  ( -POOL_SIZE,   POOL_SIZE  );

    glTexCoord2f( 1,            0          );
    glVertex2f  ( POOL_SIZE,    POOL_SIZE  );

    glTexCoord2f( 1,            1          );
    glVertex2f  ( POOL_SIZE,    -POOL_SIZE );
    glEnd();


    if( Capture )
    {
        glBindTexture(GL_TEXTURE_2D, PoolTex[DUDVMAP]);
        snap("dudv_fbo.png",POOLBUF_SIZE, POOLBUF_SIZE, TEXBUF_SRC);
    }

    glPopMatrix();
}


//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void DrawToReflectionFB()
{
    // copy the default camera to a temporary array
    float default_cam_pos[3];
    float default_cam_ori[16];
    underwater = false;

    get_camera_position( default_cam_pos, 
                         default_cam_pos + 1, 
                         default_cam_pos + 2);
    get_camera_orientation( default_cam_ori );

    if( default_cam_pos[1] < 0 ) underwater = true;

    // ===================== REFLECTION BUFFER =========================== //
    // draw to reflection buffer
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, PoolFBO[REFLECT]);
    glViewport(0,0,POOLBUF_SIZE,POOLBUF_SIZE);
    
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_FRONT);
    glDisable(GL_BLEND);
     
    glEnable(GL_CLIP_PLANE0);
    double plane[4] = { 0.0, -1.0, 0.0, 0.0 };
    glClipPlane(GL_CLIP_PLANE0, plane);

    LightPos[1] *= -1;
    glLightfv(GL_LIGHT0,GL_POSITION,LightPos);
    LightPos[1] *= -1;

    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, 4.0f / 3.0f, 0.01f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  
    // use the same one for now
    apply_camera_matrix();
    glUseProgramObjectARB(0);

    // draw
    glPushMatrix();
//    if( default_cam_pos[1] > 0 )
        glScalef( 1.0f, -1.0f, 1.0f);
    glTranslatef(0.0f, 0.5f, 0.0f);
    draw_model(poolRoomObj);
    draw_model(skySphereObj);
    draw_model(spikedTeapotObj);
    draw_model(poolBoxObj);
    glPopMatrix();

    glDisable(GL_CLIP_PLANE0);

    unapply_camera_matrix();

    if( Capture )
    {
        glBindTexture(GL_TEXTURE_2D, PoolTex[REFLECT]);
        snap("reflect_fbo.png",POOLBUF_SIZE,POOLBUF_SIZE,TEXBUF_SRC);
    }
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void DrawToRefractionFB()
{
    glEnable(GL_LIGHTING);

    // ===================== REFRACTION BUFFER =========================== //

    // draw to refraction buffer
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, PoolFBO[REFRACT]);
    glViewport(0,0,POOLBUF_SIZE,POOLBUF_SIZE);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glDisable(GL_BLEND);
    glLightfv(GL_LIGHT0,GL_POSITION,LightPos);
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, 4.0f / 3.0f, 0.01f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // same camera 
    apply_camera_matrix();
    
    glPushMatrix();
    glTranslatef(0.0f, 0.5f, 0.0f);
    // draw
    draw_model(poolRoomObj);
    draw_model(skySphereObj);
    draw_model(spikedTeapotObj);
    draw_model(poolBoxObj);
    glPopMatrix();

    unapply_camera_matrix();
    
    if( Capture )
    {
        glBindTexture(GL_TEXTURE_2D, PoolTex[REFRACT]);
        snap("refract_fbo.png",POOLBUF_SIZE, POOLBUF_SIZE, TEXBUF_SRC);
    }

}


//----------------------------------------------------------------
// 
//----------------------------------------------------------------
double underwater_plane = 0;

void DrawToDefaultFB()
{
    glActiveTexture(GL_TEXTURE1_ARB); glBindTexture(GL_TEXTURE_2D, 0);
    glActiveTexture(GL_TEXTURE0_ARB); glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glViewport(0, 0, width, height);
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glDisable(GL_BLEND);
    glEnable(GL_COLOR_MATERIAL);
    glLightfv(GL_LIGHT0,GL_POSITION,LightPos);
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
        
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0f, 4.0f / 3.0f, .01, 10000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    apply_camera_matrix();
    glPushMatrix();
    glTranslatef(0.0f, 0.5f, 0.0f);
    bind_model_shader_prg(poolRoomObj);
    if( underwater )
    {
        glEnable(GL_CLIP_PLANE0);
        GLdouble plane[4] = { 0.0, -1.0, 0.0, -0.4 };
        glClipPlane(GL_CLIP_PLANE0, plane);
        glUniform1i( get_model_uniform_loc(poolRoomObj,"Underwater"), 1);
    }
    else
    {
        glEnable(GL_CLIP_PLANE0);
        GLdouble plane[4] = { 0.0, 1.0, 0.0, 0.6 };
        glClipPlane(GL_CLIP_PLANE0, plane);
        glUniform1i( get_model_uniform_loc(poolRoomObj,"Underwater"), 0);
    }

    draw_model(poolRoomObj);
    draw_model(skySphereObj);
    draw_model(spikedTeapotObj);
    draw_model(poolBoxObj);
    glDisable(GL_CLIP_PLANE0);
    glDisable(GL_FOG);

    glPopMatrix();

    // bind textures
    glActiveTexture(GL_TEXTURE4_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolDB[REFRACT]);
    glActiveTexture(GL_TEXTURE3_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolTex[DUDVMAP]);
    glActiveTexture(GL_TEXTURE2_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolTex[NRMLMAP]);
    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolTex[REFRACT]);
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, PoolTex[REFLECT]);    

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glUseProgramObjectARB(PoolPrg);
    
    if( underwater )
        glUniform1i( glGetUniformLocationARB(PoolPrg,"UnderWater"), 1);
    else
        glUniform1i( glGetUniformLocationARB(PoolPrg,"UnderWater"), 0);

    GLfloat waterdiff[4] = { 0.5f, 0.5f, 0.5f, 0.6f };
    glMaterialfv(GL_FRONT, GL_DIFFUSE, waterdiff);

    glDisable(GL_CULL_FACE);
    glBegin(GL_QUADS);
    glBindTexture(GL_TEXTURE_2D, 0);
        glTexCoord2f(0,0);
        glVertex3f( -9.46908, 0.0, -9.46908);
        glTexCoord2f(0,1);
        glVertex3f( -9.46908, 0.0, 9.46908);
        glTexCoord2f(1,1);
        glVertex3f( 9.46908, 0.0, 9.46908);
        glTexCoord2f(1,0);
        glVertex3f( 9.46908, 0.0, -9.46908);
    glEnd();
    
    glDisable(GL_BLEND);

    unapply_camera_matrix();
    if( Capture )
    {
        Capture = false;
        snap("default_fb.png",width,height,FRAME_SRC);
    }

    SDL_GL_SwapBuffers();


}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void ProcessMouseMotion(SDL_Event* e)
{
    CamPitch += (e->motion.yrel) / float(height) * 180.0f;
    CamYaw   -= (e->motion.xrel) / float(width ) * 360.0f;
    orient_camera(DEG2RAD(CamPitch), DEG2RAD(CamYaw), 0.0f);
}

//----------------------------------------------------------------
// Process key up presses
// 
//----------------------------------------------------------------
void ProcessKeyUp(SDL_Event* e)
{
    switch( e->key.keysym.sym )
    {
    case SDLK_LEFT:
    case SDLK_RIGHT:
    case SDLK_UP:
    case SDLK_DOWN:
        LightVel[0] = LightVel[1] = LightVel[2] = 0;
        break;
    }

    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'w':
            CamVel[2] = 0;
            break;
        case 's':
            CamVel[2] = 0;
            break;
        case 'a':
            CamVel[0] = 0;
            break;
        case 'd':
            CamVel[0] = 0;
            break;
        }
    }
}

//----------------------------------------------------------------
// Process key down presses
// 
//----------------------------------------------------------------
void ProcessKeyDown(SDL_Event* e)
{
    switch( e->key.keysym.sym )
    {
    case SDLK_ESCAPE:
        BreakLoop = true;
        break;
    case SDLK_LEFT:
        LightVel[0] = -10;
        break;
    case SDLK_RIGHT:
        LightVel[0] = 10;
        break;
    case SDLK_UP:
        LightVel[2] = 10;
        break;
    case SDLK_DOWN:
        LightVel[2] = -10;
        break;
    }


    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'i':
            underwater_plane += 1;
            break;
        case 'k':
            underwater_plane -= 1;
            break;
        case 'w':
            CamVel[2] = 10;
            break;
        case 's':
            CamVel[2] = -10;
            break;
        case 'a':
            CamVel[0] = 10;
            break;
        case 'd':
            CamVel[0] = -10;
            break;
        case 'n':
            bind_model_shader_prg(poolRoomObj);
            NMapScale -= .01;
            glUniform1f( get_model_uniform_loc(poolRoomObj,"scale"),NMapScale);
            printf("Parallax Scale %f Bias %f\n", NMapScale, NMapBias);
            break;
        case 'N':
            bind_model_shader_prg(poolRoomObj);
            NMapScale += .01;
            glUniform1f( get_model_uniform_loc(poolRoomObj,"scale"),NMapScale);
            printf("Parallax Scale %f Bias %f\n", NMapScale, NMapBias);
            break;
        case 'b':
            bind_model_shader_prg(poolRoomObj);
            NMapBias -= .01;
            glUniform1f( get_model_uniform_loc(poolRoomObj,"bias"),NMapBias);
            printf("Parallax Scale %f Bias %f\n", NMapScale, NMapBias);
            break;
        case 'B':
            bind_model_shader_prg(poolRoomObj);
            NMapBias += .01;
            glUniform1f( get_model_uniform_loc(poolRoomObj,"bias"),NMapBias);
            printf("Parallax Scale %f Bias %f\n", NMapScale, NMapBias);
            break;
        case 'C':
        case 'c':
            Capture = true;
            break;
        case 'm':
            PopBackWave();
            break;
        case 'M':
            PushBackWave();
            break;
        }
    }
}


//----------------------------------------------------------------
// Infinite Loop
// 
//----------------------------------------------------------------
void InfLoop()
{
    double starttime, endtime;
    starttime = clock();
    endtime = starttime;
    int dt = 0;
    starttime = (double) SDL_GetTicks() / 1000.0; 
    
    //--- infinite loop with event queue processing
    SDL_Event event;
    while(!BreakLoop) 
    {
        while( SDL_PollEvent( &event )) 
        {
            switch( event.type )
            {
            case SDL_KEYDOWN:
                ProcessKeyDown(&event);
                break;
            case SDL_KEYUP:
                ProcessKeyUp(&event);
                break;
            case SDL_MOUSEMOTION:
                ProcessMouseMotion(&event);
                break;
            case SDL_QUIT:
                BreakLoop = true;
                break;
            }

        } // -- while event in queue

       
        //capture frame time
        endtime = (double) SDL_GetTicks() / 1000.0; 
        
        if( endtime - starttime > 1.0f / float(DESIRED_FPS) )
        {
            if( updatefn )
                updatefn(endtime - starttime);
            starttime = (double) SDL_GetTicks() / 1000.0;
            
            // update scene 
            DrawScene();
        }

        
    } // -- infinite loop
}


//----------------------------------------------------------------
// homebrew update funciton
// 
//----------------------------------------------------------------
void Update(double dt)
{
    for( int i = 0; i < 3; i++)
    {
        LightPos[i] += LightVel[i] * dt;
    }
    translate_camera( CamVel[0] * dt, CamVel[1] * dt, CamVel[2] * dt);
    TotalTime += dt;
}


//----------------------------------------------------------------
// Setup lights
// 
//----------------------------------------------------------------
void InitLight()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    
    GLfloat lightamb[] = { .3, .3, .3, 1 };
    GLfloat lightdif[] = { 1, 1, 1, 1 };
    glShadeModel(GL_SMOOTH);
    glLightfv(GL_LIGHT0,GL_AMBIENT,lightamb);
    glLightfv(GL_LIGHT0,GL_DIFFUSE,lightdif);
    glLightfv(GL_LIGHT0,GL_POSITION,LightPos);

}


//----------------------------------------------------------------
// SETUP SDL
// 
//----------------------------------------------------------------
void SetupSDL()
{
    // init video system
    const SDL_VideoInfo* vidinfo;
    if( SDL_Init(SDL_INIT_VIDEO) < 0 )
    {
        fprintf(stderr,"Failed to initialize SDL Video!\n");
        exit(1);
    }
    
    // tell system which funciton to process when exit() call is made
    atexit(SDL_Quit);

    SDL_EnableUNICODE(1);

    // get optimal video settings
    vidinfo = SDL_GetVideoInfo();
    if(!vidinfo)
    {
        fprintf(stderr,"Coudn't get video information!\n%s\n", SDL_GetError());
        exit(1);
    }

    // set opengl attributes
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE,        5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,      5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,       5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      32);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,    1);

    // get a framebuffer
    gDrawSurface = SDL_SetVideoMode(width,height,vidinfo->vfmt->BitsPerPixel,
        SDL_OPENGL);

    if( !gDrawSurface )
    {
        fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
        fprintf(stderr,"Trying lower depth size to 24\n");
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      24);
        gDrawSurface = SDL_SetVideoMode(width,height,
            vidinfo->vfmt->BitsPerPixel, SDL_OPENGL);
        if(!gDrawSurface)
        {
            fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
            fprintf(stderr,"Trying lower depth size 16\n");
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      16);
            gDrawSurface = SDL_SetVideoMode(width,height,
                vidinfo->vfmt->BitsPerPixel, SDL_OPENGL);
            if(!gDrawSurface)
            {
                fprintf(stderr,"Couldn't set video mode!\n%s\n", 
                    SDL_GetError());
                exit(1);
            }
        }
    }

    // give the window a name
    SDL_WM_SetCaption("ARUN RAO - CS594 PROJ2", "CS594 PROJ2");
    SDL_ShowCursor(0);
    SDL_WM_GrabInput(SDL_GRAB_ON);

    // set opengl viewport and perspective view
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glViewport(0,0,width,height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( 60, 4.0f / 3.0f, .01, 10000);
    glMatrixMode(GL_MODELVIEW);
}

#ifdef __APPLE__
//----------------------------------------------------------------
void SetupGLProcs()
{
}

#elif !defined(_WIN32)

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void SetupGLProcs()
{
    // SHADER PROCS
    glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glCreateProgramObjectARB");
    glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)
        glGetProcAddress("glCreateShaderObjectARB");
    glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)
        glGetProcAddress("glShaderSourceARB");
    glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)
        glGetProcAddress("glCompileShaderARB");
    glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
        glGetProcAddress("glGetObjectParameterivARB");
    glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)
        glGetProcAddress("glAttachObjectARB");
    glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)
        glGetProcAddress("glGetInfoLogARB");
    glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)
        glGetProcAddress("glLinkProgramARB");
    glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glUseProgramObjectARB");
    glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
        glGetProcAddress("glGetUniformLocationARB");
    glUniform1f = (PFNGLUNIFORM1FARBPROC)
        glGetProcAddress("glUniform1fARB");
    glUniform2f = (PFNGLUNIFORM2FARBPROC)
        glGetProcAddress("glUniform2fARB");
    glUniform3f = (PFNGLUNIFORM3FARBPROC)
        glGetProcAddress("glUniform3fARB"); 
    glUniform1fv = (PFNGLUNIFORM1FVARBPROC)
        glGetProcAddress("glUniform1fvARB");
    glUniform2fv = (PFNGLUNIFORM2FVARBPROC)
        glGetProcAddress("glUniform2fvARB");
    glUniform3fv = (PFNGLUNIFORM3FVARBPROC)
        glGetProcAddress("glUniform3fvARB"); 
    glUniform1i = (PFNGLUNIFORM1IARBPROC)
        glGetProcAddress("glUniform1iARB");
    glDeleteObjectARB = (PFNGLDELETEOBJECTARBPROC)
        glGetProcAddress("glDeleteObjectARB");
    glBindAttribLocationARB = (PFNGLBINDATTRIBLOCATIONARBPROC)
        glGetProcAddress("glBindAttribLocationARB");
    glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)
        glGetProcAddress("glGetAttribLocationARB");
    glVertexAttrib3fARB = (PFNGLVERTEXATTRIB3FARBPROC)
        glGetProcAddress("glVertexAttrib3fARB");
    

    // FBO FUNCTIONS
    glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)
        glGetProcAddress("glGenFramebuffersEXT");
    glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)
        glGetProcAddress("glBindFramebufferEXT");
    glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC)
        glGetProcAddress("glDeleteFramebuffersEXT");
    glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)
        glGetProcAddress("glFramebufferTexture2DEXT");
    glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)
        glGetProcAddress("glCheckFramebufferStatusEXT");

    // MULTITEXTURE FUNCTIONS
    glActiveTexture = (PFNGLACTIVETEXTUREPROC)
        glGetProcAddress("glActiveTexture");
}

#endif


//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void CheckGLErrors()
{
    switch( glGetError() )
    {
    case GL_NO_ERROR: break;
    case GL_INVALID_ENUM: 
        printf("GL ERR: Unacceptable value for enum argument\n");
        break;
    case GL_INVALID_VALUE: 
        printf("GL ERR: Numeric argument out of range\n");
        break;
    case GL_INVALID_OPERATION:
        printf("GL ERR: Operation not allowed in current state\n");
        break;
    case GL_STACK_OVERFLOW:
        printf("GL ERR: Command would cause stack overflow\n");
        break;
    case GL_STACK_UNDERFLOW:
        printf("GL ERR: Command would cause stack underflow\n");
        break;
    case GL_OUT_OF_MEMORY:
        printf("GL ERR: Not enough memory left to execute command\n");
        break;
    case GL_TABLE_TOO_LARGE:
        printf("GL ERR: Specified table exceeds implementations max\n");
        break;
    }
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void PrintFBOStatus()
{
    printf("FBO Status: ");
    switch( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT))
    {
    case GL_FRAMEBUFFER_COMPLETE_EXT: 
        printf("Complete!\n"); 
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
        printf("Incomplete Attachment.\n");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
        printf("Incomplte Missing Attachment.\n");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
        printf("Incomplete Problems with Dimensions.\n");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
        printf("Incomplete Problems with Attachment Formats.\n");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
        printf("Incomplete Problems with Draw Buffer.\n");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
        printf("Incomplete Problems with Read Buffer.\n");
        break;
    case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
        printf("Unsupported.  Check GL version!\n");
        break;
    }
}

//----------------------------------------------------------------
// Swap the FBO's framebuffer with another one and store the current in
// FBOFrontBuffer
//----------------------------------------------------------------
void SwapFBOBuffer(GLuint* backbuffer)
{
    if(!backbuffer) return;
    GLuint temp = *backbuffer;

    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
                              GL_COLOR_ATTACHMENT0_EXT,
                              GL_TEXTURE_2D, FBOFrontBuffer, 0);
    
    *backbuffer  = FBOFrontBuffer;
    FBOFrontBuffer = temp;

}

//=====================================================================//

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void CreatePoolFBS()
{
    glGenFramebuffersEXT(4,PoolFBO);
    glGenTextures(4,PoolTex);
    glGenTextures(2,PoolDB);
    glGenTextures(1,&FBOFrontBuffer);

    for( int i = 0; i < 4; i++)
    {
        // create depth buffer for the Reflection and Refraction only
        if( i < 2 )
        {
            glBindTexture(GL_TEXTURE_2D, PoolDB[i]);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, POOLBUF_SIZE, 
                         POOLBUF_SIZE, 0, GL_DEPTH_COMPONENT, 
                         GL_UNSIGNED_BYTE, NULL);
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        }

        // create our render textures
        glBindTexture(GL_TEXTURE_2D, PoolTex[i]);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, POOLBUF_SIZE, POOLBUF_SIZE, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, NULL);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        glBindTexture(GL_TEXTURE_2D, 0);

        //SetTexGen();
       
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, PoolFBO[i]);

        // complete attachments
        if( i < 2 )
        {
            glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
                                      GL_DEPTH_ATTACHMENT_EXT,
                                      GL_TEXTURE_2D, PoolDB[i], 0);
        }

        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                                  GL_TEXTURE_2D, PoolTex[i], 0);
 

        PrintFBOStatus();

    }

    // Make our pseudo front buffer
    glBindTexture(GL_TEXTURE_2D, FBOFrontBuffer);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, POOLBUF_SIZE, POOLBUF_SIZE, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glBindTexture(GL_TEXTURE_2D, 0);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void PushBackWave()
{
    WaveSrc src;
    srand( SDL_GetTicks() );
    src.x = rand() % int(POOL_SIZE);
    src.z = rand() % int(POOL_SIZE);
    src.x *= pow( -1.0f, rand() % 2);
    src.z *= pow( -1.0f, rand() % 2);
    src.amp_0    = AMP_RANGE * (float) rand() / (RAND_MAX + 1.0f) + AMP_MIN;
    src.lambda_0 = LAMBDA_RANGE * (float) rand() / (RAND_MAX + 1.0f) + 
        LAMBDA_MIN;
    src.start_time = SDL_GetTicks();
    src.speed = SPEED_RANGE * (float) rand() / (RAND_MAX + 1.0f) + SPEED_MIN;

    wave_vec.push_back(src);
    printf("++++ WAVE GENERATED ++++\n");
    printf("Src: %f, %f\n", wave_vec[ wave_vec.size() - 1].x,
           wave_vec[ wave_vec.size() - 1].z);
    printf("Amp: %f Lambda: %f\n", wave_vec[ wave_vec.size() -1 ].amp_0,
           wave_vec[ wave_vec.size() - 1].lambda_0);
    printf("Speed: %f\n", wave_vec[ wave_vec.size() - 1].speed);
    printf("++++++++++++++++++++++++\n");
   
    NumWaves++;

    // go through all the waves and make sure the sum of their amp
    // is not greater than .5, otherwise scale all the amps down

/*    printf("Scaling all amplitudes so sum of all will be .5\n");
    float sum = 0.0f;
    int i;
    for( i = 0; i < wave_vec.size(); i++)
    {
        sum += wave_vec[i].amp_0;
    }

    float scale = .5f / sum;
    printf("Scale factor %f\n",scale);
    for( i = 0; i < wave_vec.size(); i++)
        wave_vec[i].amp_0 *= scale;
*/
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void PopBackWave()
{
    if( wave_vec.size() > 2 )
    {   
        wave_vec.pop_back();
        printf("Popped Random Circular Wave From Listing... %d left!\n",
        wave_vec.size());
        NumWaves--;
    }
}
