#include "model.h"
#include "obj.h"
#include "matrix.h"
#include <vector>
#include <math.h>

bool is_model(int);

//------------------------------------
std::vector< ObjModel* > gModelVec;
bool ObjInitialized = false;

//------------------------------------
int load_model (const char* mfile, const char* vfile, const char* ffile) 
{
    if(!ObjInitialized) obj_add_file(NULL);

    ObjModel* mdl = new ObjModel();
    
    mdl->obj = obj_add_file(mfile);
    if( mdl->obj == 0)
    {
        delete mdl;
        return 0;
    }
    obj_proc_file(mdl->obj);    

    mdl->pos[0]   = mdl->pos[1]   = mdl->pos[2]   = 0.0f;
    mdl->scale[0] = mdl->scale[1] = mdl->scale[2] = 1.0f;
    make_identity(mdl->m);
    SetupShaders( vfile, ffile, mdl->vs, mdl->fs, mdl->prg);
    // bind the sixth index of the generic vertex attributes to the 
    // variable name of "Tangent"
    // set the shadow sampler "shmap" to 3
    // glUniform1i( glGetUniformLocationARB(mdl->prg,"shmap"), 3);
    gModelVec.push_back(mdl);
    
    return gModelVec.size() - 1;
}

//------------------------------------
int load_model (const char* mfile, int shader_copy) 
{
    if(!ObjInitialized) obj_add_file(NULL);

    ObjModel* mdl = new ObjModel();
    
    mdl->obj = obj_add_file(mfile);
    if( mdl->obj == 0)
    {
        delete mdl;
        return 0;
    }
    obj_proc_file(mdl->obj);    

    mdl->pos[0]   = mdl->pos[1]   = mdl->pos[2]   = 0.0f;
    mdl->scale[0] = mdl->scale[1] = mdl->scale[2] = 1.0f;
    make_identity(mdl->m);
    if( is_model(shader_copy))
    {
        mdl->prg = gModelVec[shader_copy]->prg;
        mdl->fs  = gModelVec[shader_copy]->fs;
        mdl->vs  = gModelVec[shader_copy]->vs;
    }
    else
    {
        mdl->prg = mdl->fs = mdl->vs = 0;
    }
    // bind the sixth index of the generic vertex attributes to the 
    // variable name of "Tangent"
    // set the shadow sampler "shmap" to 3
    // glUniform1i( glGetUniformLocationARB(mdl->prg,"shmap"), 3);
    gModelVec.push_back(mdl);
    
    return gModelVec.size() - 1;
}

//------------------------------------
void free_model(int i)
{
    if( i < 0 || i > gModelVec.size() - 1) return;
    if( gModelVec[i] )
    {
        UnloadShaders(gModelVec[i]->vs, gModelVec[i]->fs, gModelVec[i]->prg);
        obj_del_file(gModelVec[i]->obj);
        delete gModelVec[i];
        gModelVec[i] = NULL;
    }

    // move all the back ones forward one and pop the back
    for( ; i < gModelVec.size(); i++)
    {
        gModelVec[i] = gModelVec[i + 1];
    }

    gModelVec.pop_back();
}

//------------------------------------
void free_all_models()
{
    for( int i = gModelVec.size() - 1; i > 0; i--)
    {
        UnloadShaders(gModelVec[i]->vs, gModelVec[i]->fs, gModelVec[i]->prg);
        obj_del_file(gModelVec[i]->obj);
        delete gModelVec[i];
        gModelVec[i] = NULL;
    }

    obj_del_file(0);
    ObjInitialized = false;
    gModelVec.clear();
}

//------------------------------------
bool is_model(int i)
{
    if( gModelVec.size() <= 0) return false;
    if( i < 0 || i > gModelVec.size() - 1 ) return false;
    if( gModelVec[i] == NULL ) // fluke
    {
        free_model(i);
        return false;
    }
    
    return true;
}

//------------------------------------
void position_model(int i, float x, float y, float z)
{
    if( !is_model(i) ) return;
        
    gModelVec[i]->pos[0] = x;
    gModelVec[i]->pos[1] = y;
    gModelVec[i]->pos[2] = z;
}

//------------------------------------
void orient_model(int i, float p, float y, float r)
{
    if( !is_model(i) ) return;
    
    GLfloat *m = gModelVec[i]->m;
    GLfloat a[16];
    GLfloat b[16];
    make_identity(m);
    make_identity(a);
    make_identity(b);

    GLfloat cr = cosf(r);
    GLfloat sr = sinf(r);
    GLfloat cp = cosf(p);
    GLfloat sp = sinf(p);
    GLfloat cy = cosf(y);
    GLfloat sy = sinf(y);

    // pitch then yaw then roll
    a[5]  = cp;
    a[6]  = sp;
    a[9]  = -sp;
    a[10] = cp;
    b[0]  = cy;
    b[2]  = -sy;
    b[8]  = sy;
    b[10] = cy;

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);

    b[0] = cr;
    b[1] = sr;
    b[4] = -sr;
    b[5] = cr;

    mul_matrix(m,a,b);
}

//------------------------------------
void scale_model(int i, float x, float y, float z)
{
    if( !is_model(i) ) return;

    gModelVec[i]->scale[0] = x;
    gModelVec[i]->scale[1] = y;
    gModelVec[i]->scale[2] = z;
}

//------------------------------------
void scale_model(int i, float s)
{
    scale_model(i,s,s,s);
}

//------------------------------------
void translate_model(int i, float x, float y, float z)
{
    if(!is_model(i)) return;

    // orient vector based on our rotation matrix
    float tx, ty, tz;
    GLfloat *m = gModelVec[i]->m;
    tx = (m[0] * x) + (m[4] * y) + (m[8] * z);
    ty = (m[1] * x) + (m[5] * y) + (m[9] * z);
    tz = (m[2] * x) + (m[6] * y) + (m[10] * z);

    // add our new vector to our position vector
    gModelVec[i]->pos[0] += tx;
    gModelVec[i]->pos[1] += ty;
    gModelVec[i]->pos[2] += tz;
}

//------------------------------------
void turn_model(int i, float p, float y, float r)
{
    if(!is_model(i)) return;

    GLfloat *m = gModelVec[i]->m;
    GLfloat a[16];
    GLfloat b[16];
 
    copy_matrix(a,m);
    make_identity(b);

    GLfloat cr = cosf(r);
    GLfloat sr = sinf(r);
    GLfloat cp = cosf(p);
    GLfloat sp = sinf(p);
    GLfloat cy = cosf(y);
    GLfloat sy = sinf(y);

    // pitch then yaw then roll
    b[5]  = cp;
    b[6]  = sp;
    b[9]  = -sp;
    b[10] = cp;
    
//    printf("before transforming\n");
//    print_matrix(m);

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);
//    printf("after pitch\n");
//    print_matrix(m);

    b[0]  = cy;
    b[2]  = -sy;
    b[8]  = sy;
    b[10] = cy;

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);
//    printf("after yaw\n");
//    print_matrix(m);

    b[0] = cr;
    b[1] = sr;
    b[4] = -sr;
    b[5] = cr;

    mul_matrix(m,a,b);
//    printf("after roll\n");
//    print_matrix(m);

}

//------------------------------------
void stretch_model(int i, float x, float y, float z)
{
    if(!is_model(i)) return;

    gModelVec[i]->scale[0] *= x;
    gModelVec[i]->scale[1] *= y;
    gModelVec[i]->scale[2] *= z;
}

//------------------------------------
void stretch_model(int i, float s)
{
    stretch_model(i,s,s,s);
}

//------------------------------------
void draw_model(int i, bool shader, bool axes)
{
    if(!is_model(i)) return;
    GLfloat *p = gModelVec[i]->pos;
    GLfloat *r = gModelVec[i]->m;
    GLfloat *s = gModelVec[i]->scale;

    // bind the program object
    if( shader)
        glUseProgramObjectARB(gModelVec[i]->prg);
    else
        glUseProgramObjectARB(0);

    glPushMatrix();
        glTranslatef(-p[0],-p[1],-p[2]);
        glMultMatrixf(r);
        glScalef(s[0],s[1],s[2]);
        obj_draw_file(gModelVec[i]->obj);
        if( axes )
            obj_draw_axes(gModelVec[i]->obj,1.0);
    glPopMatrix();

}

//------------------------------------
GLint get_model_uniform_loc(int i, const char* label)
{
    if(!is_model(i)) return -1;
    return glGetUniformLocationARB( gModelVec[i]->prg, label);
}

void bind_model_shader_prg(int i)
{
    if(!is_model(i)) 
        glUseProgramObjectARB(0);
    else
        glUseProgramObjectARB(gModelVec[i]->prg);
}

void set_models_do_lighting(bool b)
{
    int v = b ? 1 : 0;
    for( int i = 0; i < gModelVec.size(); ++i)
    {
       glUniform1i(glGetUniformLocationARB(gModelVec[i]->prg, "DoLighting"),v);
    }
}

